from pathlib import Path
from typing import List

import aiofiles
import aiohttp
from bs4 import BeautifulSoup

from .characterfactory import Character
from .extractor import Extractor


class FileIconsExtractor(Extractor):
    __icons: List[Character]

    def __init__(self):
        self.__icons = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_icons()
        await self.__write_to_file(target)
        print("Finished File Icons")

    async def __fetch_icons(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://raw.githubusercontent.com/file-icons/icons/refs/heads/master/charmap.md"
            ) as response:
                for line in (await response.text()).split("\n"):
                    if line and line.strip().startswith("<tbody>"):
                        html = BeautifulSoup(line, "lxml")
                        name = html.find_all("td")[1].string
                        codepoint = int(html.find_all("td")[2].code.string[1:], 16)
                        self.__icons.append(
                            Character(
                                codepoint,
                                name,
                            )
                        )

    async def __write_to_file(self, target: Path) -> None:
        async with aiofiles.open(target / "fileicons.csv", mode="w") as character_file:
            for icon in self.__icons:
                await character_file.write(f"{icon.directional_char} {icon.lower_case_name}\n")



from subprocess import run
from typing import Dict, List, Tuple, Union

from ..abstractionhelper import is_installed
from ..models import CANCEL, DEFAULT, Action, CharacterEntry, Shortcut
from .selector import Selector


class Rofi(Selector):
    @staticmethod
    def supported() -> bool:
        return is_installed("rofi")

    @staticmethod
    def name() -> str:
        return "rofi"

    def show_character_selection(
        self,
        characters: List[CharacterEntry],
        recent_characters: List[str],
        prompt: str,
        show_description: bool,
        use_icons: bool,
        keybindings: Dict[Action, str],
        additional_args: List[str],
    ) -> Tuple[Union[Action, DEFAULT, CANCEL], Union[List[str], Shortcut]]:
        parameters = [
            "rofi",
            "-dmenu",
            "-markup-rows",
            "-i",
            "-multi-select",
            "-no-custom",
            "-ballot-unselected-str",
            "",
            "-format",
            "i",
            "-p",
            prompt,
            "-kb-custom-11",
            keybindings[Action.COPY],
            "-kb-custom-12",
            keybindings[Action.TYPE],
            "-kb-custom-13",
            keybindings[Action.CLIPBOARD],
            "-kb-custom-14",
            keybindings[Action.TYPE_NUMERICAL],
            "-kb-custom-15",
            keybindings[Action.UNICODE],
            "-kb-custom-16",
            keybindings[Action.COPY_UNICODE],
            *additional_args,
        ]

        if recent_characters:
            parameters.extend(["-mesg", self.__format_recent_characters(recent_characters)])

        rofi = run(
            parameters,
            input="\n".join(self.__format_characters(characters, use_icons, show_description)),
            capture_output=True,
            encoding="utf-8",
        )

        if 10 <= rofi.returncode <= 19:
            return DEFAULT(), Shortcut(rofi.returncode - 10)

        action: Union[Action, DEFAULT, CANCEL]
        if rofi.returncode == 1:
            action = CANCEL()
        elif rofi.returncode == 20:
            action = Action.COPY
        elif rofi.returncode == 21:
            action = Action.TYPE
        elif rofi.returncode == 22:
            action = Action.CLIPBOARD
        elif rofi.returncode == 23:
            action = Action.UNICODE
        elif rofi.returncode == 24:
            action = Action.COPY_UNICODE
        elif rofi.returncode == 25:
            action = Action.TYPE_NUMERICAL
        else:
            action = DEFAULT()

        return action, [characters[int(index)].character for index in rofi.stdout.splitlines()]

    def __format_characters(
        self, characters: List[CharacterEntry], use_icons: bool, show_description: bool
    ) -> List[str]:
        if use_icons and not show_description:
            return [f"\0meta\x1f{entry.description}\x1ficon\x1f<span>{entry.character}</span>" for entry in characters]
        elif use_icons and show_description:
            return [f"{entry.description}\0icon\x1f<span>{entry.character}</span>" for entry in characters]
        elif not use_icons and show_description:
            return self.basic_format_characters(characters, strip_tags=False)
        else:
            return [f"{entry.character}\0meta\x1f{entry.description}" for entry in characters]

    def __format_recent_characters(self, recent_characters: List[str]) -> str:
        pairings = [f"\u200e{(index + 1) % 10}: {character}" for index, character in enumerate(recent_characters)]

        return " | ".join(pairings)

    def show_skin_tone_selection(
        self, tones_emojis: List[str], prompt: str, additional_args: List[str]
    ) -> Tuple[int, str]:
        rofi = run(
            ["rofi", "-dmenu", "-i", "-no-custom", "-p", prompt, *additional_args],
            input="\n".join(tones_emojis),
            capture_output=True,
            encoding="utf-8",
        )

        return rofi.returncode, rofi.stdout

    def show_action_menu(self, additional_args: List[str]) -> List[Action]:
        rofi = run(
            [
                "rofi",
                "-dmenu",
                "-multi-select",
                "-no-custom",
                "-ballot-unselected-str",
                "",
                "-i",
                *additional_args,
            ],
            input="\n".join([str(it) for it in Action if it != Action.MENU]),
            capture_output=True,
            encoding="utf-8",
        )

        return [Action(action) for action in rofi.stdout.strip().split("\n")]



import asyncio
import pathlib

from .blockextractor import BlockExtractor
from .characterfactory import CharacterFactory
from .cjkextractor import CjkExtractor
from .emojiextractor import EmojiExtractor
from .extractor import Extractor
from .fileiconsextractor import FileIconsExtractor
from .fontawesome6extractor import FontAwesome6Extractor
from .gitmojiextractor import GitmojiExtractor
from .htmlcharacterextractor import HtmlCharacterExtractor
from .kaomojiextractor import KaomojiExtractor
from .mathcollectionextractor import MathExtractor
from .nerdfontextractor import NerdFontExtractor
from .weathericonsextractor import WeatherIconsExtractor


async def extract_all():
    data_directory = pathlib.Path(__file__).parent.parent / "picker" / "data"

    character_factory = CharacterFactory()
    await character_factory.fetch_characters()

    extractors = [
        EmojiExtractor(),
        BlockExtractor(character_factory),
        CjkExtractor(),
        MathExtractor(character_factory),
        NerdFontExtractor(),
        GitmojiExtractor(),
        FontAwesome6Extractor(),
        KaomojiExtractor(),
        HtmlCharacterExtractor(),
        FileIconsExtractor(),
        WeatherIconsExtractor(),
    ]

    for subclass in Extractor.__subclasses__():
        if not any(isinstance(extractor, subclass) for extractor in extractors):
            print(f"Did you forget to add an extractor of class {subclass.__name__}?")

    await asyncio.gather(*[extractor.extract_to(data_directory) for extractor in extractors])


if __name__ == "__main__":
    asyncio.run(extract_all())



from pathlib import Path
from typing import List

import aiofiles
import aiohttp
from bs4 import BeautifulSoup

from .characterfactory import Character
from .extractor import Extractor


class WeatherIconsExtractor(Extractor):
    __icons: List[Character]

    def __init__(self):
        self.__icons = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_icons()
        await self.__write_to_file(target)
        print("Finished Weather Icons")

    async def __fetch_icons(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://raw.githubusercontent.com/erikflowers/weather-icons/refs/heads/master/values/weathericons.xml"
            ) as response:
                html_content = BeautifulSoup(await response.text(), "xml")

                for row in html_content.find("resources").find_all("string"):
                    self.__icons.append(Character(row.string, row["name"]))

    async def __write_to_file(self, target: Path) -> None:
        async with aiofiles.open(target / "weathericons.csv", mode="w") as character_file:
            for icon in self.__icons:
                await character_file.write(f"{icon.directional_char} {icon.lower_case_name}\n")



from subprocess import run
from typing import Dict, List, Tuple, Union

from ..abstractionhelper import is_installed, is_wayland
from ..models import CANCEL, DEFAULT, Action, CharacterEntry, Shortcut
from .selector import Selector


class Fuzzel(Selector):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("fuzzel")

    @staticmethod
    def name() -> str:
        return "fuzzel"

    def show_character_selection(
        self,
        characters: List[CharacterEntry],
        recent_characters: List[str],
        prompt: str,
        show_description: bool,
        use_icons: bool,
        keybindings: Dict[Action, str],
        additional_args: List[str],
    ) -> Tuple[Union[Action, DEFAULT, CANCEL], Union[List[str], Shortcut]]:
        parameters = ["fuzzel", "--dmenu", "--fuzzy-min-length", "1", "--index", "-p", prompt, *additional_args]

        fuzzel = run(
            parameters, input="\n".join(self.basic_format_characters(characters)), capture_output=True, encoding="utf-8"
        )
        return DEFAULT(), [characters[int(fuzzel.stdout.strip())].character]

    def show_skin_tone_selection(
        self, tones_emojis: List[str], prompt: str, additional_args: List[str]
    ) -> Tuple[int, str]:
        fuzzel = run(
            ["fuzzel", "--dmenu", "--fuzzy-min-length", "1", "-p", prompt, *additional_args],
            input="\n".join(tones_emojis),
            capture_output=True,
            encoding="utf-8",
        )

        return fuzzel.returncode, fuzzel.stdout

    def show_action_menu(self, additional_args: List[str]) -> List[Action]:
        fuzzel = run(
            [
                "fuzzel",
                "--dmenu",
                *additional_args,
            ],
            input="\n".join([it.value for it in Action if it != Action.MENU]),
            capture_output=True,
            encoding="utf-8",
        )

        return [Action(fuzzel.stdout.strip())]



import html
from typing import Dict, List, Optional, Union
from unicodedata import bidirectional

import aiohttp


class Character:
    ltr_mark: str = "\u200e"

    char: str
    name: str
    force_ltr: bool
    descriptions: List[str]

    def __init__(
        self,
        char: Union[int, str],
        name: str,
        bidi_class: Optional[str] = None,
        descriptions: Optional[List[str]] = None,
    ):
        self.char = chr(char) if isinstance(char, int) else char
        self.name = name.strip()
        if not bidi_class:
            bidi_class = bidirectional(self.char)
        self.force_ltr = bidi_class in ("AL", "AN", "R", "RLE", "RLI", "RLO")
        if descriptions:
            self.descriptions = descriptions
        else:
            self.descriptions = []

    @property
    def directional_char(self) -> str:
        return f"{self.ltr_mark if self.force_ltr else ''}{self.char}"

    @property
    def title_case_name(self) -> str:
        return html.escape(self.name.title())

    @property
    def lower_case_name(self) -> str:
        return html.escape(self.name.lower())

    def add_description(self, description: str) -> None:
        if description not in self.descriptions and description != self.name and description != self.char:
            self.descriptions.append(description)

    def add_descriptions(self, descriptions: List[str]) -> None:
        for description in descriptions:
            self.add_description(description)


class CharacterFactory:
    __characters: Dict[str, Character]

    def __init__(self):
        self.__characters = {}

    async def fetch_characters(self) -> None:
        INDEX_CODEPOINT = 0
        INDEX_NAME = 1
        INDEX_CATEGORY = 2
        INDEX_BIDI_CLASS = 4

        async with aiohttp.ClientSession() as session:
            async with session.get("https://unicode.org/Public/UNIDATA/UnicodeData.txt") as response:
                lines = (await response.text()).split("\n")

                for line in lines:
                    fields = line.split(";")
                    if (
                        len(fields) >= 2
                        and fields[INDEX_CATEGORY] not in ("Cc", "Co", "Cs")
                        and not (fields[INDEX_NAME].startswith("<") and fields[INDEX_NAME].endswith(">"))
                    ):
                        character = Character(
                            int(fields[INDEX_CODEPOINT], 16), fields[INDEX_NAME], fields[INDEX_BIDI_CLASS]
                        )
                        self.__characters[character.char] = character

    def get_character(self, char: int) -> Optional[Character]:
        return self.__characters.get(chr(char))



from pathlib import Path
from typing import List

import aiofiles
import aiohttp

from .characterfactory import Character, CharacterFactory
from .extractor import Extractor


class MathExtractor(Extractor):
    __char_factory: CharacterFactory
    __characters: List[Character]

    def __init__(self, character_factory: CharacterFactory):
        self.__char_factory = character_factory
        self.__characters = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_math_symbols()
        await self.__write_file(target)
        print("Finished Math characters")

    async def __fetch_math_symbols(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get("https://unicode.org/Public/math/latest/MathClassEx-15.txt") as response:
                characters = []
                for line in (await response.text()).split("\n"):
                    if line and not line.startswith("#"):
                        fields = line.split(";")
                        symbols = self.__resolve_character_range(fields[0].strip())
                        for symbol in symbols:
                            characters.append(self.__char_factory.get_character(symbol))

                self.__characters = [character for character in characters if character is not None]

    def __resolve_character_range(self, line: str) -> List[int]:
        try:
            (start, end) = line.split("..")
            return list(range(int(start, 16), int(end, 16) + 1))
        except ValueError:
            return [int(line, 16)]

    async def __write_file(self, target: Path) -> None:
        async with aiofiles.open(target / "math.csv", mode="w") as character_file:
            for character in self.__characters:
                await character_file.write(f"{character.char} {character.title_case_name}\n")



import html
from dataclasses import dataclass
from pathlib import Path
from typing import List

import aiofiles
import aiohttp

from .extractor import Extractor


@dataclass
class Kaomoji:
    string: str
    description: str


class KaomojiExtractor(Extractor):
    __kaomojis: List[Kaomoji]

    def __init__(self):
        self.__kaomojis = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_icons()
        await self.__write_to_file(target)
        print("Finished Kaomoji")

    async def __fetch_icons(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://raw.githubusercontent.com/w33ble/emoticon-data/master/emoticons.json"
            ) as response:
                """
                {
                  "tags": [
                    "hide"
                  ],
                  "string": "(⊃‿⊂)",
                  "id": "f73741a5-13f9-44a3-9dfe-4bce15c44cda"
                }
                """
                for kaomoji in (await response.json(content_type=None))["emoticons"]:
                    self.__kaomojis.append(Kaomoji(kaomoji["string"].replace(" ", " "), ", ".join(kaomoji["tags"])))

    async def __write_to_file(self, target: Path) -> None:
        async with aiofiles.open(target / "kaomoji.csv", mode="w") as character_file:
            for kaomoji in self.__kaomojis:
                await character_file.write(f"{kaomoji.string} {html.escape(kaomoji.description.lower())}\n")



from dataclasses import dataclass
from pathlib import Path
from typing import List

import aiofiles
import aiohttp

from .extractor import Extractor


@dataclass
class HtmlCharacter:
    string: str
    name: str


class HtmlCharacterExtractor(Extractor):
    __characters: List[HtmlCharacter]

    def __init__(self):
        self.__characters = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_data()
        await self.__write_to_file(target)
        print("Finished HtmlCharacters")

    async def __fetch_data(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get("https://html.spec.whatwg.org/entities.json") as response:
                entities = await response.json()
                for key in entities:
                    if key.endswith(";") and key != "&NewLine;":
                        name = key[1:-1]
                        char = "".join(list(map(chr, entities[key]["codepoints"])))
                        self.__characters.append(HtmlCharacter(char, name))

    async def __write_to_file(self, target: Path) -> None:
        async with aiofiles.open(target / "html_characters.csv", mode="w") as character_file:
            for character in self.__characters:
                await character_file.write(f"{character.string} {character.name}\n")



from subprocess import run
from typing import List

from ..abstractionhelper import is_installed
from .typer import Typer


class CliclickTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_installed("cliclick")

    @staticmethod
    def name() -> str:
        return "cliclick"

    def get_active_window(self) -> str:
        return ""

    def type_characters(self, characters: str, active_window: str) -> None:
        pass


    def insert_from_clipboard(self, active_window: str) -> None:
        run(
            [
                "cliclick",
                "-w 100",
                "kd:cmd",
                "t:v",
                "ku:cmd"
            ]
        )

    def type_numerical(self, codepoints: List[int], active_window: str) -> None:
        pass



import asyncio
import io
import re
import zipfile
from pathlib import Path
from typing import Dict, List

import aiofiles
import aiohttp

from .characterfactory import Character
from .extractor import Extractor


class CjkExtractor(Extractor):
    __characters: Dict[str, List[Character]]

    def __init__(self):
        self.__characters = {}

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_characters()

        languages = ("Cantonese", "Mandarin", "Vietnamese", "Tang", "JapaneseKun", "JapaneseOn", "Korean")
        await asyncio.gather(
            *[self.__write_to_file(target, language, self.__characters[language]) for language in languages]
        )
        print("Finished CJK")

    async def __fetch_characters(self) -> None:
        INDEX_CODEPOINT = 0
        INDEX_LANGUAGE = 1
        INDEX_DESCRIPTION = 2

        async with aiohttp.ClientSession() as session:
            async with session.get("https://unicode.org/Public/UNIDATA/Unihan.zip") as response:
                characters: Dict[str, List[Character]] = {}
                with zipfile.ZipFile(io.BytesIO(await response.read())) as zip:
                    with io.TextIOWrapper(zip.open("Unihan_Readings.txt"), encoding="utf-8") as file:
                        for line in file.read().split("\n"):
                            if not line or line.startswith("#"):
                                continue

                            fields = line.split("\t")
                            character = Character(int(fields[INDEX_CODEPOINT][2:], 16), fields[INDEX_DESCRIPTION], "L")
                            language = fields[INDEX_LANGUAGE][1:]
                            if language not in characters:
                                characters[language] = []
                            characters[language].append(character)

                self.__characters = characters

    async def __write_to_file(self, target: Path, language: str, characters: List[Character]) -> None:
        filename = f"cjk_{re.sub(r'(?!^)(?=[A-Z])', '_', language).lower()}.csv"
        async with aiofiles.open(target / filename, mode="w") as character_file:
            for character in characters:
                await character_file.write(f"{character.directional_char} {character.title_case_name}\n")



from dataclasses import dataclass
from typing import List

from .characterfactory import Character, CharacterFactory


@dataclass
class Block:
    name: str
    characters: List[Character]


class BlockFactory:
    __char_factory: CharacterFactory

    def __init__(self, character_factory: CharacterFactory):
        self.__char_factory = character_factory

    def build_block_from_range(self, name: str, range: str) -> Block:
        (start, end) = range.split("..")
        return self.__build_block(name, int(start.strip(), 16), int(end.strip(), 16))

    def __build_block(self, name: str, start: int, end: int) -> Block:
        return Block(name.strip().title(), self.__fill_characters(start, end))

    def __fill_characters(self, start: int, end: int) -> List[Character]:
        return [
            char
            for char in (self.__char_factory.get_character(pos) for pos in range(start, end + 1))
            if char is not None
        ]



from pathlib import Path
from typing import List

import aiofiles
import aiohttp

from .characterfactory import Character
from .extractor import Extractor


class NerdFontExtractor(Extractor):
    __icons: List[Character]

    def __init__(self):
        self.__icons = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_icons()
        await self.__write_to_file(target)
        print("Finished NerdFont")

    async def __fetch_icons(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://raw.githubusercontent.com/ryanoasis/nerd-fonts/master/glyphnames.json"
            ) as response:
                for key, value in (await response.json(content_type=None)).items():
                    if key == "METADATA":
                        continue
                    icon = int(value["code"], 16)
                    self.__icons.append(Character(icon, key))

    async def __write_to_file(self, target: Path) -> None:
        async with aiofiles.open(target / "nerd_font.csv", mode="w") as character_file:
            for icon in self.__icons:
                await character_file.write(f"{icon.directional_char} {icon.lower_case_name}\n")



from pathlib import Path
from typing import List

import aiofiles
import aiohttp

from .characterfactory import Character
from .extractor import Extractor


class GitmojiExtractor(Extractor):
    __icons: List[Character]

    def __init__(self):
        self.__icons = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_icons()
        await self.__write_to_file(target)
        print("Finished Gitmoji")

    async def __fetch_icons(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://raw.githubusercontent.com/carloscuesta/gitmoji/master/packages/gitmojis/src/gitmojis.json"
            ) as response:
                """
                {
                    'emoji': '🎨',
                    'entity': '&#x1f3a8;',
                    'code': ':art:',
                    'description': 'Improve structure / format of the code.',
                    'name': 'art',
                    'semver': None
                }
                """
                for char in (await response.json(content_type=None))["gitmojis"]:
                    icon = ord(char["emoji"][0])
                    name = char["description"]
                    self.__icons.append(Character(icon, name))

    async def __write_to_file(self, target: Path) -> None:
        async with aiofiles.open(target / "gitmoji.csv", mode="w") as character_file:
            for icon in self.__icons:
                await character_file.write(f"{icon.directional_char} {icon.lower_case_name}\n")



import typing
from pathlib import Path

import aiofiles
import aiohttp

from .blockfactory import Block, BlockFactory
from .characterfactory import CharacterFactory
from .extractor import Extractor


class BlockExtractor(Extractor):
    __blocks: typing.List[Block]
    __block_factory: BlockFactory

    def __init__(self, character_factory: CharacterFactory):
        super().__init__()
        self.__blocks = []
        self.__block_factory = BlockFactory(character_factory)

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_blocks()
        await self.__write_to_files(target)
        print("Finished Unicode")

    async def __fetch_blocks(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get("https://www.unicode.org/Public/17.0.0/ucd/Blocks.txt") as response:
                lines = (await response.text()).split("\n")

                for line in lines:
                    if not line or line.startswith(("#", "@")):
                        continue
                    key, _, value = line.partition(";")
                    self.__blocks.append(self.__block_factory.build_block_from_range(value.strip(), key.strip()))

    async def __write_to_files(self, target: Path) -> None:
        for block in self.__blocks:
            if not block.characters:
                continue

            async with aiofiles.open(
                target / f"{block.name.lower().replace(' ', '_')}.csv", mode="w"
            ) as character_file:
                for character in block.characters:
                    await character_file.write(f"{character.directional_char} {character.title_case_name}\n")



from pathlib import Path
from typing import List

import aiofiles
import aiohttp

from .characterfactory import Character
from .extractor import Extractor


class FontAwesome6Extractor(Extractor):
    __icons: List[Character]

    def __init__(self):
        self.__icons = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_icons()
        await self.__write_to_file(target)
        print("Finished FontAwesome 6")

    async def __fetch_icons(self) -> None:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                "https://raw.githubusercontent.com/FortAwesome/Font-Awesome/6.x/metadata/icons.json"
            ) as response:
                for name, data in (await response.json(content_type=None)).items():
                    icon = int(data["unicode"], 16)
                    name = f"fa-{name}"
                    char = Character(icon, name)
                    try:
                        for alias in data["aliases"]["names"]:
                            char.add_description(f"fa-{alias}")
                    except KeyError:
                        pass
                    self.__icons.append(char)

    async def __write_to_file(self, target: Path) -> None:
        async with aiofiles.open(target / "fontawesome6.csv", mode="w") as character_file:
            for icon in self.__icons:
                aliases = f" <small>({', '.join(icon.descriptions)})</small>"
                line = f"{icon.directional_char} {icon.lower_case_name}{aliases if icon.descriptions else ''}\n"
                await character_file.write(line)



import asyncio
import html
from pathlib import Path
from typing import Dict, List

import aiofiles
import aiohttp
from aiohttp import ClientSession, ClientTimeout
from bs4 import BeautifulSoup
from lxml import etree
from lxml.etree import XPath

from .blockfactory import Block
from .characterfactory import Character
from .extractor import Extractor


class EmojiExtractor(Extractor):
    __all_blocks: List[Block]
    __annotations: Dict[str, List[str]]
    __base_emojis: List[str]

    def __init__(self):
        self.__annotations = {}
        self.__all_blocks = []
        self.__ep_emojis = []
        self.__base_emojis = []

    async def extract_to(self, target: Path) -> None:
        await self.__fetch_data()
        await self.__write_character_file(target)
        await self.__write_metadata_file(target)
        print("Finished Emoji")

    async def __fetch_data(self) -> None:
        async with aiohttp.ClientSession(timeout=ClientTimeout(sock_read=120)) as session:
            await asyncio.gather(
                self.__fetch_emoji_list(session),
                self.__fetch_annotations(session),
                self.__fetch_additional_data(session),
            )

    async def __fetch_emoji_list(self, session: ClientSession) -> None:
        async with session.get("https://unicode.org/emoji/charts-17.0/full-emoji-list.html") as data:
            html_content = BeautifulSoup(await data.text(), "lxml")

            current_title = None
            current_emojis: List[Character] = []
            for row in html_content.find("table").find_all("tr"):
                if row.th and "bighead" in row.th["class"]:
                    if current_title:
                        self.__all_blocks.append(Block(current_title, current_emojis))
                    current_title = row.a.string
                    current_emojis = []
                elif not row.th:
                    emoji = row.find("td", {"class": "chars"}).string
                    description = row.find("td", {"class": "name"}).string.replace("⊛ ", "")
                    current_emojis.append(
                        Character(
                            emoji if emoji in self.__ep_emojis or len(emoji) > 1 else f"{emoji}️",
                            description,
                            "L",
                            self.__annotations.get(emoji, None),
                        )
                    )

            self.__all_blocks.append(Block(current_title, current_emojis))

    async def __fetch_annotations(self, session: ClientSession) -> None:
        async with session.get(
            "https://raw.githubusercontent.com/unicode-org/cldr/latest/common/annotations/en.xml"
        ) as data:
            xpath = XPath('./annotations/annotation[not(@type="tts")]')
            for element in xpath(etree.fromstring(await data.read())):
                self.__annotations[element.get("cp")] = element.text.split(" | ")

    async def __fetch_additional_data(self, session: ClientSession) -> None:
        def __extract_ep_emojis(emoji_data: List[str]) -> None:
            started = False
            emojis = []
            for line in emoji_data:
                if not started and not line.startswith("# All omitted code points have Emoji_Presentation=No"):
                    continue
                else:
                    started = True
                if line.startswith("# Total elements:"):
                    break
                if line and not line.startswith("#"):
                    emojis.extend(self.__resolve_character_range(line.split(";")[0].strip()))

            self.__ep_emojis = emojis

        def __extract_base_emojis(emoji_data: List[str]) -> None:
            started = False
            emojis = []
            for line in emoji_data:
                if not started and not line.startswith("# All omitted code points have Emoji_Modifier_Base=No"):
                    continue
                else:
                    started = True
                if line.startswith("# Total elements:"):
                    break
                if line and not line.startswith("#"):
                    emojis.extend(self.__resolve_character_range(line.split(";")[0].strip()))

            self.__base_emojis = emojis

        async with session.get("https://unicode.org/Public/17.0.0/ucd/emoji/emoji-data.txt") as data:
            emoji_data = (await data.text()).split("\n")
            __extract_ep_emojis(emoji_data)
            __extract_base_emojis(emoji_data)

    def __resolve_character_range(self, line: str) -> List[str]:
        try:
            (start, end) = line.split("..")
            return [chr(char) for char in range(int(start, 16), int(end, 16) + 1)]
        except ValueError:
            return [self.__resolve_character(line)]

    def __resolve_character(self, string: str) -> str:
        return "".join(chr(int(character, 16)) for character in string.split(" "))

    async def __write_character_file(self, target: Path) -> None:
        for block in self.__all_blocks:
            async with aiofiles.open(
                target / f"emojis_{block.name.lower().replace(' ', '').replace('&', '_')}.csv", mode="w"
            ) as character_file:
                for entry in self.__compile_entries(block):
                    await character_file.write(entry + "\n")

    def __compile_entries(self, block: Block) -> List[str]:
        annotated_emojis = []
        for emoji in block.characters:
            entry = f"{emoji.char} {html.escape(emoji.name)}"
            if emoji.descriptions:
                entry += f" <small>({html.escape(', '.join(emoji.descriptions))})</small>"
            annotated_emojis.append(entry)
        return annotated_emojis

    async def __write_metadata_file(self, target: Path) -> None:
        async with aiofiles.open(target / "copyme.py", mode="w") as metadata_file:
            await metadata_file.write("skin_tone_selectable_emojis={'")
            await metadata_file.write("', '".join(self.__base_emojis))
            await metadata_file.write("'}\n")
